import numpy as np

__all__ = ["ecdf"]


def ecdf(samples):
    """Empirical Cumulative Distribution Function
    acts like a proxy for the Cumulative distribution
    function and can be used to predict the central
    tendency of a random sample.

    Parameters
    ----
    samples: array_like
             Observed Data.
    
    Returns
    -----
    x_ecdf, y_ecdf: array_like
                    Sorted samples x_ecdf and corresponding
                    quantiles y_ecdf.

    Examples
    -----
    >>> from swissroll.utils import ecdf
    >>> import numpy as np
    >>> sample = np.random.randn(1000)
    >>> x_ecdf, y_ecdf = ecdf(sample)
    """
    x_ecdf = np.sort(samples)
    y_ecdf = np.arange(1, len(x_ecdf) + 1) / len(x_ecdf)

    return x_ecdf, y_ecdf



import numpy as np
from scipy.stats import norm, t
from scipy.special import gamma, factorial
import matplotlib.pyplot as plt
from ._distribution import Discrete, Continuous

__all__ = [
    "Bernoulli",
    "Binomial",
    "Beta",
    "Normal",
    "Cauchy",
    "Exponential",
    "Poisson",
    "StudentT",
    "Uniform",
]


class Bernoulli(Discrete):
    """Bernoulli distribution.

    Parameters
    -----
    p: (np.float32, np.float64)
        Probability of success.
    
    Examples
    -----
    >>> from swissroll.distributions import Bernoulli
    >>> dist = Bernoulli(0.3)
    >>> dist.pmf(0.)
    0.7
    >>> print(dist)
    Bernoulli distribution with parameters {'p': 0.3}
    >>> dist.sample(3)
    np.array([1., 0., 1.])
    """

    def __init__(self, p=0.5):
        self.p = p

    def pmf(self, x, p=None):
        if p is not None:
            return p ** x * (1.0 - p) ** (1.0 - x)
        if x not in [0, 1]:
            raise ValueError(
                f"Support for Bernoulli distribution" " is {0, 1}." f" found: {x}."
            )

        return self.p ** x * (1.0 - self.p) ** (1.0 - x)

    def cdf(self, x):
        if x < 0.0:
            return 0.0
        if x < 1.0:
            return 1.0 - self.p
        return 1.0

    def meanvar(self):
        return self.p, self.p * (1.0 - self.p)

    def sample(self, size=10000):
        return np.random.binomial(1.0, self.p, size=size)

    def plot_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=1.0, num=1000)
        for i in theta:
            plt.scatter(i, np.prod(self.pmf(observed, i)), color="r")
        if show:
            plt.show()

    def plot_log_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=1.0, num=1000)
        for i in theta:
            plt.scatter(i, np.sum(np.log(self.pmf(observed, i))), color="r")
        if show:
            plt.show()

    def mle(self, observed):
        return np.mean(observed)


class Binomial(Discrete):
    def __init__(self, n, p):
        self.n = n
        self.p = p

    def pmf(self, x, n=None, p=None):
        if n and p:
            return (
                factorial(n)
                / (factorial(x) * factorial(n - x))
                * (p ** x)
                * ((1.0 - p) ** (n - x))
            )
        return (
            factorial(self.n)
            / (factorial(x) * factorial(self.n - x))
            * (self.p ** x)
            * ((1.0 - self.p) ** (self.n - x))
        )

    def cdf(self, x):
        res = np.zeros_like(x)
        for i, j in enumerate(x):
            res[i] += np.sum(self.pmf(np.arange(0, j, dtype=np.float32)))

        return res

    def meanvar(self):
        return self.n * self.p, self.n * self.p * (1.0 - self.p)

    def sample(self, size=10000):
        return np.random.binomial(self.n, self.p, size=size)

    def plot_likelihood(self, n, observed, show=False):
        theta = np.linspace(start=0.0, stop=1.0, num=1000)
        for i in theta:
            plt.scatter(i, np.prod(self.pmf(observed, n, i)), color="r")
        if show:
            plt.show()

    def plot_log_likelihood(self, n, observed, show=False):
        theta = np.linspace(start=0.0, stop=1.0, num=1000)
        for i in theta:
            plt.scatter(i, np.sum(np.log(self.pmf(observed, n, i))), color="r")
        if show:
            plt.show()

    def mle(self, n, observed):
        return np.mean(observed)


class Poisson(Discrete):
    def __init__(self, lambda_):
        self.lambda_ = lambda_

    def pmf(self, x, lambda_=None):
        if lambda_ is not None:
            return np.exp(-lambda_) * (lambda_ ** x) / factorial(x)
        return np.exp(-self.lambda_) * (self.lambda_ ** x) / factorial(x)

    def cdf(self, x):
        res = np.zeros_like(x)
        for i, j in enumerate(x):
            res[i] += np.sum(self.pmf(np.arange(0, j, dtype=np.float32)))

        return res

    def meanvar(self):
        return self.lambda_, self.lambda_

    def sample(self, size=10000):
        return np.random.poisson(self.lambda_, size=size)

    def plot_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=2 * observed.mean(), num=1000)
        for i in theta:
            plt.scatter(i, np.prod(self.pmf(observed, i)), color="r")
        if show:
            plt.show()

    def plot_log_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=2 * observed.mean(), num=1000)
        for i in theta:
            plt.scatter(i, np.sum(np.log(self.pmf(observed, i))), color="r")
        if show:
            plt.show()

    def mle(self, observed):
        return np.mean(observed)


class Exponential(Continuous):
    def __init__(self, lambda_):
        self.lambda_ = lambda_

    def pdf(self, x, lambda_=None):
        if lambda_ is not None:
            return (x >= 0.0) * lambda_ * np.exp(-lambda_ * x)
        return (x >= 0.0) * self.lambda_ * np.exp(-self.lambda_ * x)

    def cdf(self, x):
        return (x >= 0.0) * (1.0 - np.exp(-self.lambda_ * x))

    def meanvar(self):
        return 1.0 / self.lambda_, 1.0 / (self.lambda_ ** 2)

    def sample(self, size=10000):
        return np.random.exponential(self.lambda_, size=size)

    def plot_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=2.0 / observed.mean(), num=1000)
        for i in theta:
            plt.scatter(i, np.prod(self.pdf(observed, i)), color="r")
        if show:
            plt.show()

    def plot_log_likelihood(self, observed, show=False):
        theta = np.linspace(start=0.0, stop=2.0 / observed.mean(), num=1000)
        for i in theta:
            plt.scatter(i, np.sum(np.log(self.pdf(observed, i))), color="r")
        if show:
            plt.show()

    def mle(self, observed):
        import warnings

        if not np.mean(observed):
            warnings.warn(
                f"mean of the data is zero"
                f" whih is not valid for exponential"
                f"distribution.",
                RuntimeWarning,
            )
        return 1.0 / (np.mean(observed) + 1e-9)


class Normal(Continuous):
    def __init__(self, mu, var):
        self.mu = mu
        self.var = var

    def pdf(self, x, mu=None, var=None):
        if mu and var:
            return (1.0 / np.sqrt(2 * np.pi * var)) * np.exp(-0.5 * (x - mu) ** 2 / var)
        return (1.0 / np.sqrt(2 * np.pi * self.var)) * np.exp(
            -0.5 * (x - self.mu) ** 2 / self.var
        )

    def cdf(self, x):
        return norm.cdf(x, loc=self.mu, scale=self.var)

    def sample(self, size=10000):
        return np.random.normal(self.mu, self.var, size=size)

    def meanvar(self):
        return self.mu, self.var

    def plot_likelihood(self, observed, mu=None, var=None, show=False):
        if mu is None and var is None:
            raise ValueError(
                "at least one parameter must be known" " to plot the likelihood!"
            )
        if mu is None:
            theta = np.linspace(
                start=-2 * observed.mean() - 2,
                stop=2 * observed.mean() + 2,
                num=1000,
                dtype=np.float32,
            )
            for i in theta:
                plt.scatter(i, np.prod(self.pdf(observed, i, var)), color="r")
            if show:
                plt.show()
        elif var is None:
            theta = np.linspace(
                start=-2 * np.var(observed),
                stop=2 * np.var(observed),
                num=1000,
                dtype=np.float32,
            )
            for i in theta:
                plt.scatter(i, np.prod(self.pdf(observed, mu, i)), color="r")

            if show:
                plt.show()

    def plot_log_likelihood(self, observed, mu=None, var=None, show=False):
        if mu is None and var is None:
            raise ValueError(
                "at least one parameter must be known" " to plot the likelihood!"
            )
        if mu is None:
            theta = np.linspace(
                start=-2 * observed.mean() - 2,
                stop=2 * observed.mean() + 2,
                num=1000,
                dtype=np.float32,
            )
            for i in theta:
                plt.scatter(i, np.sum(np.log(self.pdf(observed, i, var))), color="r")
            if show:
                plt.show()
        elif var is None:
            theta = np.linspace(
                start=0.0, stop=2 * np.var(observed), num=1000, dtype=np.float32
            )
            for i in theta:
                plt.scatter(i, np.sum(np.log(self.pdf(observed, mu, i))), color="r")
            if show:
                plt.show()

    def mle(self, observed, mu=None, var=None):
        if mu is None and var is None:
            raise ValueError("at least one parameter must be known" " to find the mle!")

        if mu is None:
            return np.mean(observed)
        elif var is None:
            return np.sum((observed - mu) ** 2) / len(observed)


class Uniform(Continuous):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def pdf(self, x):
        return (self.a <= x) * (x <= self.b) / (self.b - self.a)

    def cdf(self, x):
        return np.array(
            list(
                map(
                    lambda i: (i >= self.a) * (i - self.a) / (self.b - self.a)
                    if i <= self.b
                    else 1.0,
                    x,
                )
            )
        )

    def sample(self, size=10000):
        return np.random.uniform(self.a, self.b, size=size)

    def meanvar(self):
        return (self.a + self.b) / 2.0, (self.b - self.a) ** 2 / 2

    def plot_likelihood(self, observed):
        pass

    def plot_log_likelihood(self, observed):
        pass

    def mle(self, observed):
        return np.max(observed)


class Beta(Continuous):
    def __init__(self, alpha, beta):
        self.alpha = alpha
        self.beta = beta

    def pdf(self, x):
        x = np.array(x, dtype=np.float64)
        return (
            gamma(self.alpha + self.beta)
            / (gamma(self.alpha) * gamma(self.beta))
            * x ** (self.alpha - 1)
            * (1 - x) ** (self.beta - 1)
        )

    def cdf(self, x):
        pass


class StudentT(Continuous):
    def __init__(self, dof):
        self.dof = dof

    def pdf(self, x):
        return (
            gamma((self.dof + 1.0) / 2)
            / (np.sqrt(self.dof * np.pi) * gamma(self.dof / 2.0))
            * (1.0 + x ** 2 / self.dof) ** (-(self.dof + 1) / 2)
        )

    def cdf(self, x):
        return t.cdf(x, df=self.dof)

    def sample(self, size=10000):
        return t.rvs(df=self.dof, size=size)

    def plot_likelihood(self, observed):
        pass

    def plot_log_likelihood(self, observed):
        pass

    def mle(self, observed):
        pass


class Cauchy(Continuous):
    def __init__(self, x0, gamma):
        self.x0 = x0
        self.gamma = gamma

    def pdf(self, x):
        return (self.gamma / np.pi) / ((x - self.x0) ** 2 + self.gamma ** 2)

    def cdf(self, x):
        return 0.5 + (1.0 / np.pi) * np.arctan((x - self.x0) / self.gamma)

    def meanvar(self):
        return np.nan, np.nan

    def sample(self):
        pass

    def plot_likelihood(self, observed):
        pass

    def plot_log_likelihood(self, observed):
        pass

    def mle(self, observed):
        pass



import numpy as np

__all__ = ["GPRegressor"]


def rbf_kernel(x1, x2, sigma, landmark):
    """Evaluate the rbf kernel

    Parameters
    -----
    x1: matrix of shape `(n_samples, n_features)`

    x2: vector of size `(n_features)`

    sigma: scalar

    landmark: scalar

    Returns
    -----
    rbf: vector of size `(n_samples)`
    """
    x2 = x2.reshape(1, -1)
    if x1.shape[1] != x2.shape[1]:
        raise ValueError(
            f"shapes of x1 and x2 are inconsistent => "
            f"{x1.shape[1]} != {x2.shape[1]}"
        )
    s = sigma
    l = landmark
    return s ** 2.0 * np.exp(
        -0.5 * np.linalg.norm(x1 - x2, axis=-1) ** 2.0 / (l ** 2.0)
    )


def normalize(xs):
    """Center the data to match the standard scale

    Parameters
    -----
    xs: matrix of shape `(n_samples, n_features)`

    Returns
    -----
    norm_xs: matrix of shape `(n_samples, n_features)`
    """
    return (xs - np.mean(xs)) / np.std(xs)


class GPRegressor:
    """Gaussian Process Regressor

    Parameters
    -----
    k_sigma: scalar
        standard deviation of RBF kernel

    k_landmark: scalar
        landmark of the RBF kernel
    """

    def __init__(self, k_sigma, k_landmark):
        self.k_sigma = k_sigma
        self.k_landmark = k_landmark
        self.__fitted = False

    def fit(self, xs, ys, verbose=False):
        """Fit the regressor on your dataset

        Parameters
        -----
        xs: matrix of shape `(n_samples, n_features)`
            training data to be fitted

        ys: vector of size `(n_samples)`
            target values

        verbose: bool, optional
            if `True`, prints the results after calculation
        """
        self.n_samples = xs.shape[0]
        self.n_features = xs.shape[1]
        self._x = xs
        self._y = ys

        self.mus = np.zeros_like(ys.reshape(-1, 1))
        self.sigmas = np.zeros((self.n_samples, self.n_samples))
        for i in range(self.n_samples):
            self.sigmas[i, :] = rbf_kernel(xs, xs[i], self.k_sigma, self.k_landmark)

        if verbose:
            print(self.mus)
            print(self.sigmas)
            print(self.mus.shape)
            print(self.sigmas.shape)

        self.__fitted = True

    def predict(self, xs, predict_uncertainties=False):
        """Predict on the new datapoint

        Parameters
        -----
        x: matrix of shape `(n_new_samples, n_features)`
            new datapoints for which the target is to be predicted

        predict_uncertainties: bool, optional
            if `True`, return the covarience matrix with predicted target values

        Returns
        -----
        pos_mus: vector of size `(n_new_samples)`
            predicted target values
        
        pos_sigmas: (if `predict_uncertainties=True`) matrix of shape (n_new_samples, n_new_samples)
            covarience matrix of the predicted targets
        """
        if not self.__fitted:
            raise TypeError(f"no dataset fitted! First call the `fit` method")
        if xs.shape[1] != self.n_features:
            raise ValueError(
                f"inconsistent input to method `predict` "
                f"{xs.shape[1]} != {self.n_features}"
            )

        k = np.zeros((xs.shape[0], self.n_samples))
        for i in range(xs.shape[0]):
            k[i, :] = rbf_kernel(self._x, xs[i], self.k_sigma, self.k_landmark)

        # pos_mus = (self.n_new_samples, 1)
        pos_mus = k @ np.linalg.inv(self.sigmas) @ self._y

        if predict_uncertainties:
            # pos_sigmas = (n_new_samples, n_new_samples)
            pos_sigmas = self.k_sigma ** 2.0 - k @ np.linalg.inv(self.sigmas) @ k.T
            return pos_mus, pos_sigmas

        return pos_mus



from abc import ABCMeta, abstractmethod


class Distribution:
    """Abstract base class for all the distributions.

    Parameters
    -------
    **kwargs:
              Keyword arguments, if any.
    
    Attributes
    -------
    get_params: `callable`
                Get a copy of parameters of the distribution.
    
    set_params: `callable`
                Update the parameters of the distribution.
    """

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def get_params(self):
        """Get a clone of paramters of the distribution"""
        return self.__dict__.copy()

    def set_params(self, **kwargs):
        """Update the parameters of the distribution"""
        self.__init__(**kwargs)

    def __repr__(self):
        """Returns the memory address of the distribution"""
        return f"<{self.__class__.__name__} at {id(self)}>"

    def __str__(self):
        """Pretty print the distribution and its parameters"""
        message = (
            f"{self.__class__.__name__} distribution with"
            f" parameters {self.__dict__}."
        )

        return message

    def __getstate__(self):
        """Makes the class and subclasses serailizable."""
        return self.__dict__.copy()

    def __setstate__(self, state):
        """Makes the class and subclasses serializable."""
        self.__dict__.update(state)

    @property
    def _short_str(self):
        """A short string of the parameters of the distribution"""
        params = self.get_params()
        msg = ", ".join(f"{name}={val}" for name, val in params.items())
        return msg


class Continuous(Distribution, metaclass=ABCMeta):
    """Abstract base class for continuous distributions."""

    @abstractmethod
    def pdf(self, x, *args, **kwargs):
        """Evaluate the Probability Density Function of a distribution."""
        pass

    @abstractmethod
    def cdf(self, x, *args, **kwargs):
        """Cumulative Distribution Function"""
        pass

    @abstractmethod
    def plot_likelihood(self, *args, **kwargs):
        """Plot the likelihood function.
        NOTE: Use with caution! It may not work
        for a large sample size. Instead, use the 
        alternative `plot_log_likelihood`.
        """
        pass

    @abstractmethod
    def plot_log_likelihood(self, observed, *args, **kwargs):
        """Plot the log-likelihood"""
        pass

    @abstractmethod
    def mle(self, observed, *args, **kwargs):
        """Maximum Likelihood Estimation"""
        pass


class Discrete(Distribution, metaclass=ABCMeta):
    """Abstract base class for Discrete distributions."""

    @abstractmethod
    def pmf(self, x, *args, **kwargs):
        """Evaluate the Probability Mass Function of the distribution"""
        pass

    @abstractmethod
    def cdf(self, x, *args, **kwargs):
        """Evaluate Cumulative Distribution function of the distribution"""
        pass

    @abstractmethod
    def plot_likelihood(self, *args, **kwargs):
        """Plot the likelihood function.
        NOTE: Use with caution! It may not work
        for a large sample size. Instead, use the 
        alternative `plot_log_likelihood`.
        """
        pass

    @abstractmethod
    def plot_log_likelihood(self, observed, *args, **kwargs):
        """Plot the log-likelihood"""
        pass

    @abstractmethod
    def mle(self, observed, *args, **kwargs):
        """Compute Maximum Likelihood Estimation of paramters
        of the distribution."""
        pass



import warnings
import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from scipy.stats import multivariate_normal as Normal
from sklearn.datasets import make_spd_matrix

__all__ = ["GaussianMM"]


def make_simplex(n_clusters):
    """Generate a (n_clusters - 1)-D random simplex"""
    x = np.random.randn(n_clusters)
    return np.exp(x) / np.sum(np.exp(x))


class GaussianMM:
    """Gaussian Mixture Model. A soft clustering
    method using EM Algorithm.

    Parameters
    -----
    n_clusters: int, optional
                number of gaussian clusters to use.

    Examples
    -----
    >>> from swissroll.gmm import GaussianMM
    >>> model = GaussianMM(n_clusters=2)
    >>> loss, mu, sigma, pi = model.fit(X, max_iter=25, n_repeats=100, verbose=True)
    """

    def __init__(self, n_clusters=3):
        self.n_clusters = n_clusters
        self.n_samples = None
        self.n_features = None
        # we may later need these to run the e-step.
        # NOTE: all the class variables MUST be declared
        # here but can be initialized outside!
        self.posterior = None
        self.vlb = -np.inf
        self.mu = None
        self.sigma = None
        self.pi = None

    def __init_params(self, n_samples, n_features):
        """Reinitialize the parameters.

        So, here's our parameters:
        ===========================================================================
        Parameters
        -----
        n_samples : number of data points in the dataset used during fitting.
        n_features : dimentions of the dataset used during fitting.
        pi : weights of `n_clusters` gaussians aka a categorical distribution over
             latent variable.
        mu : centers of `n_clusters` gaussians aka mean of the gaussians
        sigma : shape parameter of the `n_clusters` gaussians aka covarience
                matrix of the gaussians.
        posterior : posterior distribution over the latent varables.
        ===========================================================================
        Shapes
        -----
        n_samples : scalar
        n_features : scalar
        pi : Row Vector of shape `(n_clusters, 1)`
        mu : Matrix of shape `(n_clusters, n_features)`
        sigma : Tensor of shape `(n_clusters, n_features, n_features)`
        posterior : Row vector of shape `(n_clusters, n_samples)`
        ===========================================================================
        """
        # initialize the scalars
        self.n_samples = n_samples
        self.n_features = n_features

        # initialize the `theta` parameters
        self.pi = make_simplex(self.n_clusters).reshape(-1, 1)
        self.mu = np.random.randn(self.n_clusters, self.n_features)
        self.sigma = np.asarray(
            [make_spd_matrix(self.n_features) for _ in range(self.n_clusters)]
        )

        # Initialize the posterior as zeros
        self.posterior = np.zeros((self.n_clusters, self.n_samples), dtype=np.float64)

        return None

    def _e_step(self, X):
        """Calculate the posterior which will
        minimize the gab between VLB and log-likelihood.
        Keywords: Variational Lower Bound (VLB), log-likelihood
        =====================================================
        $$\min _{\theta} \left[ \log(P(X|\theta)) - L(\theta^k, q) \right]$$
        $$L(\theta^k, q) = \sum_{i=0}^{N} \sum_{c=1}^{C} q(t_i=c) \log\left(\frac{P(x_i, t_i=c|\theta^k_c)}{q(t_i=c)}\right)$$

        which is equivalent to minimizing $KL(q(t_i=c) || P(t_i=c/x_i, \theta^k_c))$.

        Hence, we just need to set the latent distribution $q$
        equal to the posterior $P(t_i|x_i, \theta^k_c)$ in the E-Step.
        =====================================================
        Shapes
        -----
        posterior : Matrix of shape `(n_clusters, n_samples)`
        =====================================================
        """
        # First, calculate the normalization constant of the posterior.
        # NOTE: It is tractable to calculate the porterior as the latent
        # variable is finitely discrete.
        _normalization_constant = 0.0
        for cluster in range(self.n_clusters):
            _normalization_constant += (
                Normal(self.mu[cluster], self.sigma[cluster]).pdf(X) * self.pi[cluster]
            )

        # Now, we calculate the posterior for each cluster.
        for cluster in range(self.n_clusters):
            self.posterior[cluster] = (
                Normal(self.mu[cluster], self.sigma[cluster]).pdf(X)
                * self.pi[cluster]
                / _normalization_constant
            )

        return None

    def _m_step(self, X):
        """Maximization Step in EM Algorithm. Computes the
        parameters analytically that maximize the VLB.
        ==================================================
        In M-Step we maximize:
        $$L(\theta, q^k) = \sum _{i=1} ^{N} \sum _{c=1} ^{C} q^k(t_i=c)\log\left(\frac{P(x_i, t_i=c|\theta_c)}{q^k(t_i=c)}\right)$$

        This function can be diffrentiatied w.r.t to our parameters
        to get the following equations:

        Mean : $$\mu_c = \frac{\sum_{i=1}^{N} q(t_i=c)x_i}{\sum_{i=1}^{N} q(t_i=c)}$$

        Varience : $$\Sigma_c = \frac{\sum_{i=1}^{N} q(t_i=c)(x_i-\mu_c)(x_i-\mu_c)^T}{\sum_{i=1}^{N} q(t_i=c)}$$

        Weights : $$\pi_c = \frac{\sum_{i=1}^{N}q(t_i=c)}{N}$$
        ==================================================
        """
        # First, update the centers of clusters
        for cluster in range(self.n_clusters):
            self.mu[cluster] = np.nan_to_num(
                np.sum(self.posterior[cluster] * X.T, axis=1)
                / np.sum(self.posterior[cluster]),
                nan=1.0,
                posinf=1.0,
                neginf=1.0,
            )

        for cluster in range(self.n_clusters):
            self.sigma[cluster] = np.nan_to_num(
                np.sum(
                    [
                        self.posterior[cluster, i]
                        * np.outer(X[i] - self.mu[cluster], X[i] - self.mu[cluster])
                        for i in range(self.n_samples)
                    ],
                    axis=0,
                )
                / np.sum(self.posterior[cluster]),
                nan=1.0,
                posinf=1.0,
                neginf=1.0,
            )

        self.pi = np.sum(self.posterior, axis=1) / self.n_samples

        return None

    def _step(self, X):
        """Take one step of EM Algorithm"""
        # E-Step : Calculate the posterior with thetas fixed
        self._e_step(X)
        # M-Step : Calculate the values of new parameters by
        # differentiating the VLB.
        self._m_step(X)

    def _init_animate(self):
        """Initialize the animation"""
        self.ax.clear()
        self.__init_params(*self.X.shape)

        labels = self.posterior.argmax(axis=0)

        zm = self.pi[0] * Normal.pdf(self.grid, mean=self.mu[0], cov=self.sigma[0])
        for cluster in range(1, self.n_clusters):
            zm += self.pi[cluster] * Normal.pdf(
                self.grid, mean=self.mu[cluster], cov=self.sigma[cluster]
            )

        self.cf = self.ax.contourf(self.xm, self.ym, zm, alpha=0.6)
        self.sc = self.ax.scatter(
            self.X[:, 0], self.X[:, 1], c=self.colors[labels], s=30
        )
        return self.cf, self.sc

    def _fit_animate(self, frame):
        """Runs an animation while training!"""
        # This is the optional part used for plotting
        # so I can play animations. Not needed to be
        # tested! Enjoy!
        # Restart the animation every max_iter steps
        if ((frame + 1) % self.__max_iter) == 0:
            self.__init_params(*self.X.shape)

        self.ax.clear()
        try:
            if np.abs(1.0 - np.sum(self.pi)) > 1e-9:
                plt.title("Model Collapsed!")
                return self.cf, self.sc
            self._step(self.X)
            vlb = self._vlb(self.X)
            if vlb > self.vlb:
                self.vlb = vlb
        except np.linalg.LinAlgError:
            self.__init_params(*self.X.shape)
            plt.title("Model Collapsed!")
            return self.cf, self.sc

        labels = self.posterior.argmax(axis=0)

        zm = self.pi[0] * Normal.pdf(self.grid, mean=self.mu[0], cov=self.sigma[0])
        for cluster in range(1, self.n_clusters):
            zm += self.pi[cluster] * Normal.pdf(
                self.grid, mean=self.mu[cluster], cov=self.sigma[cluster]
            )

        self.cf = self.ax.contourf(self.xm, self.ym, zm, alpha=0.6)
        self.sc = self.ax.scatter(
            self.X[:, 0], self.X[:, 1], c=self.colors[labels], s=30
        )
        plt.title(f"loss : {vlb:.2f}, best_loss : {self.vlb:.2f}")
        return self.cf, self.sc

    def set_data(self, X):
        """Used only to run animations on particular datasets."""
        self.X = X

    def _vlb(self, X):
        """Evaluates the Variational Lower Bound
        for current values of parameters.
        =========================================
        $$L(\theta, q) = \sum _{i=0} ^{N} \sum _{c=1} ^{C} q(t_i=c) \log \left( P(x_i, t_i=c|\theta_c) \right)$$

        Our objective is to maximize this lower bound. The value of
        VLB is logged on the terminal to make sure that it is increasing.
        NOTE: If the value decreases, then there is something wrong with
        the algorithm. It can be proved that this value should always increase
        after each step.
        =========================================
        Shapes
        -----
        vlb : scalar
        =========================================
        """
        vlb = np.sum(
            [
                self.posterior[cluster] * np.log(self.pi[cluster])
                + Normal(self.mu[cluster], self.sigma[cluster]).logpdf(X)
                for cluster in range(self.n_clusters)
            ]
        ) - np.sum(self.posterior * np.log(self.posterior))

        return vlb

    def fit(self, X, max_iter=100, n_repeats=100, rtol=1e-9, verbose=False):
        """Fit a dataset to the model.

        Parameters
        -----
        X : array_like of shape (n_samples, n_features)
            dataset
        
        max_iter : int, optional
                   maximum iterations of the EM-Algorithm.
        
        n_repeats : int, optional
                    The number of times to repeat the algorithm
                    for better convergence.

        rtol : float, optional
               tolerance for divergence of weights of gaussians.
        
        verbose : bool, optional
                  If True, it prints the loss and best
                  loss on the terminal.
        
        Returns
        -----
        vlb, mu, sigma, pi : best loss and parameters of the model.

        Examples
        -----
        >>> from swissroll.gmm import GaussianMM
        >>> model = GaussianMM(n_clusters=3)
        >>> loss, mu, sigma, pi = model.fit(X_train)
        """
        # start training
        vlb = -np.inf
        best_mu = None
        best_sigma = None
        best_pi = None
        for _ in range(n_repeats):
            # NOTE: we are performing high dimentional
            # optimization which can often lead to divergence
            # of our constraints. In such a case, we simply
            # restart the training.
            try:
                # Initialize the parameters
                self.__init_params(*X.shape)
                # Take `max_iter` steps.
                for __ in range(max_iter):
                    # If the constraint over weights diverge
                    # restart training.
                    if np.abs(1.0 - np.sum(self.pi)) > rtol:
                        warnings.warn("Weights Diverged!", RuntimeWarning)
                        break
                    self._step(X)
                    vlb = self._vlb(X)
                    if verbose:
                        sys.stdout.write(
                            f"\rTrial {_+1}/{n_repeats}, Epoch {__+1}/{max_iter} : "
                            f"loss : {vlb:.2f}, best_loss : {self.vlb:.2f}"
                        )

                # We chosse the best parameters
                # that maximize our lower bound
                if vlb > self.vlb:
                    self.vlb = vlb
                    best_mu = self.mu
                    best_sigma = self.sigma
                    best_pi = self.pi
            except np.linalg.LinAlgError:
                warnings.warn(
                    f"Trial {_+1}: Singular Matrix: Components Collapsed",
                    RuntimeWarning,
                )
                continue

        # update the parameters according
        # to the best found so far.
        self.mu = best_mu
        self.sigma = best_sigma
        self.pi = best_pi

        return self.vlb, self.mu, self.sigma, self.pi

    def fit_animate(
        self, X, colors, frames=50, interval=100, max_iter=100, *, render_as_mp4=True
    ):
        """Visualize the training of GMMs by
        running an animation in real time!!!

        Parameters
        -----
        X : array_like
            Dataset to fit
        
        colors : array_like
                 Array of colors to assign to
                 the clusters in the animation
        
        frames : int, optional
                 number of frames in the animation.
                 only necessary while saving animation

        interval : int, optional
                   interval between two updates in the animation

        max_iter : int, optional
                   Maximum iterations of EM-Step for each
                   random initialization
        
        render_as_mp4 : bool, optional
                        If true animation is saved in the 
                        directory running the script.
        
        Examples
        -----
        >>> from swissroll.gmm import GaussianMM
        >>> model = GaussianMM(n_clusters=3)
        >>> model.fit_animate(X_train, colors=[(1, 0, 0), (0, 1, 0), (0, 0, 1)])
        """
        self.__max_iter = max_iter
        self.set_data(X)
        self.fig, self.ax = plt.subplots()
        xmin, ymin, xmax, ymax = (
            np.min(self.X[:, 0]),
            np.min(self.X[:, 1]),
            np.max(self.X[:, 0]),
            np.max(self.X[:, 1]),
        )
        xs = np.linspace(xmin - 1.0, xmax + 1.0, num=500)
        ys = np.linspace(ymin - 1.0, ymax + 1.0, num=500)
        self.xm, self.ym = np.meshgrid(xs, ys)
        self.grid = np.empty((500, 500, 2))
        self.grid[:, :, 0] = self.xm
        self.grid[:, :, 1] = self.ym
        self.colors = colors
        anim = FuncAnimation(
            self.fig,
            self._fit_animate,
            init_func=self._init_animate,
            frames=frames,
            interval=interval,
            blit=False,
        )
        if render_as_mp4:
            anim.save("gmm2d_animation.mp4")
        else:
            plt.show()

        return None

    def predict(self, X):
        """Predict the labels of a new dataset.
        
        Parameters
        -----
        X : array_like
            dataset
        
        Returns
        -----
        labels : labels corresponding to the dataset.

        Examples
        -----
        >>> from swissroll.gmm import GaussianMM
        >>> model = GaussianMM(n_clusters=3)
        >>> model.fit(X_train)
        >>> labels = model.predict(X_test)
        """
        self.posterior_predictive = self._e_step(X)
        labels = np.argmax(self.posterior_predictive, axis=0)
        return labels

    def get_params(self):
        """Get a copy of parameters of the model
        which is a tuple (mu, sigma, pi)."""
        return self.mu.copy(), self.sigma.copy(), self.pi.copy()



import collections.abc

import numpy as np
import matplotlib.pyplot as plt

from ._distribution import Discrete, Continuous
from .utils import ecdf

__all__ = ["plot_ecdf", "plot_hist", "plot_pmf", "plot_pdf", "plot_cdf"]


def plot_ecdf(samples, show=False):
    """Plot the empirical cumulative distribution function
    of the samples. This can be used to estimate the centeral
    tendency of sampled data.

    Parameters
    -----
    samples: array_like
             Observed data.
    
    Examples
    -----
    >>> import swissroll.distributions as srd
    >>> from swissroll import plotter
    >>> samples = srd.Binomial(20, 0.7)
    >>> plotter.plot_ecdf(samples)
    """
    x_ecdf, y_ecdf = ecdf(samples)
    plt.plot(x_ecdf, y_ecdf, marker=".", linestyle="none")
    if show:
        plt.show()


def plot_hist(samples, show=False):
    """Plot the histogram of the observed data.

    Parameters
    -----
    samples: array_like
             Observed data.
    
    Examples
    -----
    >>> import swissroll.distributions as srd
    >>> from swissroll import plotter
    >>> samples = srd.Binomial(20, 0.7)
    >>> plotter.plot_hist(samples)
    """
    plt.hist(samples, bins=10, density=True)
    if show:
        plt.show()


def plot_pmf(distributions, support, show=False):
    """Plot the Probability Mass Function of multiple
    distributions.

    Parameters
    -----
    distributions: list, tuple
                   distributions whose pmf is to be evaluated.
    
    support: list, array_like
             values at which pmf is to be evaluated. It can be
             different or common for all the distributions.
    
    Examples
    -----
    >>> import swissroll.distributions as srd
    >>> from swissroll import plotter
    >>> dist = srd.Binomial(20, 0.7)
    >>> support = np.arange(0, 20, dtype=np.float32)
    >>> plotter.plot_pmf(dist, support)
    """
    if isinstance(support, np.ndarray) or not isinstance(
        support, collections.abc.Collection
    ):
        plot_pmf(distributions, [support] * len(distributions))
        return

    if len(distributions) != len(support):
        raise ValueError("length of support must be equal" " to distributions!")

    for distribution, y in zip(distributions, support):
        if isinstance(distribution, Continuous):
            raise ValueError(
                "Continuous distributions don't have"
                " a mass function! Try swissroll.plotter.plot_pdf"
            )
        plt.plot(y, distribution.pmf(y), label=distribution._short_str)
    plt.legend()
    if show:
        plt.show()


def plot_pdf(distributions, support, show=False):
    """Plot the Probability Density Function of multiple
    or single distributions.

    Parameters
    -----
    distributions: list, tuple, array_like
                   distributions whose pdf is to be evaluated.
    
    support: list, tuple, array_like
             values at which pdf is to be evaluated. It can be
             different or common for all the distributions.
    
    Examples
    -----
    >>> import swissroll.distributions as srd
    >>> from swissroll import plotter
    >>> dist = srd.Binomial(20, 0.7)
    >>> support = np.arange(0, 20, dtype=np.float32)
    >>> plotter.plot_pdf(dist, support)
    """
    if isinstance(support, np.ndarray) or not isinstance(
        support, collections.abc.Collection
    ):
        plot_pdf(distributions, [support] * len(distributions))
        return

    if len(distributions) != len(support):
        raise ValueError("length of support must be equal" " to distributions!")

    for distribution, y in zip(distributions, support):
        if isinstance(distribution, Discrete):
            raise ValueError(
                "Discrete distributions don't have"
                " a density function!"
                " Try `swissroll.plotter.plot_pmf`"
            )
        plt.plot(y, distribution.pdf(y), label=distribution._short_str)
    plt.legend()
    if show:
        plt.show()


def plot_cdf(distributions, support, show=False):
    """Plot the Cumulative Distribution Function of multiple
    distributions.

    Parameters
    -----
    distributions: list, tuple
                   distributions whose cdf is to be evaluated.
    
    support: list, array_like
             values at which cdf is to be evaluated. It can be
             different for all the distributions or common.
    
    Examples
    -----
    >>> import swissroll.distributions as srd
    >>> from swissroll import plotter
    >>> dist = srd.Binomial(20, 0.7)
    >>> support = np.arange(0, 20, dtype=np.float32)
    >>> plotter.plot_cdf(dist, support)
    """
    if isinstance(support, np.ndarray) or not isinstance(
        support, collections.abc.Collection
    ):
        plot_cdf(distributions, [support] * len(distributions))
        return

    if len(distributions) != len(support):
        raise ValueError("length of support must be equal" " to distributions!")

    for distribution, y in zip(distributions, support):
        plt.plot(y, distribution.cdf(y), label=distribution._short_str)
    plt.legend()
    if show:
        plt.show()


def say_what(just_for_fun=True):
    # undocumented easter egg!
    print("What!!!")



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="swissroll",
    version="1.0.4",
    description="Higher Level API for working with Probability Distributions.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/tirthasheshpatel/swissroll",
    author="Tirth Patel and GitHub Community",
    author_email="tirthasheshpatel@gmail.com",
    license="BSD-3-Clause",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["numpy", "scipy", "matplotlib", "sklearn"],
)
